/* adapted for jack and LADCCA by Bob Ham */

/* miniFMsynth 1.0 by Matthias Nagorni    */
/* This program uses callback-based audio */
/* playback as proposed by Paul Davis on  */
/* the linux-audio-dev mailinglist.       */

#define _GNU_SOURCE

#include <stdio.h>   
#include <stdlib.h>
#include <alsa/asoundlib.h>
#include <math.h>
#include <jack/jack.h>
#include <alloca.h>

#ifdef HAVE_LADCCA
#include <ladcca/ladcca.h>
#endif

#define POLY 10

snd_seq_t *seq_handle;
double phi[POLY], phi_mod[POLY], pitch, modulation, velocity[POLY],
       attack, decay, sustain, release, env_time[POLY], env_level[POLY], gain, sample_rate;
int harmonic, subharmonic, transpose, note[POLY], gate[POLY], note_active[POLY];

int process (jack_nframes_t nframes, void * data);
int sample_rate_cb (jack_nframes_t sr, void * data);

#ifdef HAVE_LADCCA
cca_client_t * cca_client = NULL;
#endif

jack_client_t * jack_client = NULL;
jack_port_t * jack_port = NULL;

char jack_client_name[512];
char alsa_client_name[512];

snd_seq_t * open_seq () {
#ifdef HAVE_LADCCA
    cca_event_t * event;
    char id[2];
#endif
    snd_seq_t *seq_handle;
    
    if (snd_seq_open(&seq_handle, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
        fprintf(stderr, "Error opening ALSA sequencer.\n");
        exit(1);
    }
    snd_seq_set_client_name(seq_handle, alsa_client_name);
    if (snd_seq_create_simple_port(seq_handle, "miniFMsynth",
        SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_APPLICATION) < 0) {
        fprintf(stderr, "Error creating sequencer port.\n");
        exit(1);
    }

#ifdef HAVE_LADCCA
    event = cca_event_new_with_type (CCA_Alsa_Client_ID);
    id[0] = snd_seq_client_id (seq_handle);
    id[1] = '\0';
    cca_event_set_string (event, id);
    cca_send_event (cca_client, event);
#endif

    return(seq_handle);
}

void
open_jack ()
{
#ifdef HAVE_LADCCA
  cca_event_t * event;
#endif

  jack_client = jack_client_new (jack_client_name);
  if (!jack_client)
    {
      fprintf (stderr, "%s: could not connect to jack server, exiting\n",
               __FUNCTION__);
      exit (1);
    }
  
  jack_port = jack_port_register (jack_client, "out_1", JACK_DEFAULT_AUDIO_TYPE,
                                  JackPortIsOutput|JackPortIsTerminal, 0);
  if (!jack_port)
    {
      fprintf (stderr, "%s: could not register jack port, exiting\n",
               __FUNCTION__);
      exit (1);
    }
  
  jack_set_process_callback (jack_client, process, NULL);
  
  sample_rate = jack_get_sample_rate (jack_client);
  jack_set_sample_rate_callback (jack_client, sample_rate_cb, NULL);

#ifdef HAVE_LADCCA  
  event = cca_event_new_with_type (CCA_Jack_Client_Name);
  cca_event_set_string (event, jack_client_name);
  cca_send_event (cca_client, event);
#endif

  jack_activate (jack_client);
}

void
close_jack ()
{
  jack_deactivate (jack_client);
}

/*snd_pcm_t *open_pcm(char *pcm_name) {

    snd_pcm_t *playback_handle;
    snd_pcm_hw_params_t *hw_params;
    snd_pcm_sw_params_t *sw_params;
            
    if (snd_pcm_open (&playback_handle, pcm_name, SND_PCM_STREAM_PLAYBACK, 0) < 0) {
        fprintf (stderr, "cannot open audio device %s\n", pcm_name);
        exit (1);
    }
    snd_pcm_hw_params_alloca(&hw_params);
    snd_pcm_hw_params_any(playback_handle, hw_params);
    snd_pcm_hw_params_set_access(playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED);
    snd_pcm_hw_params_set_format(playback_handle, hw_params, SND_PCM_FORMAT_S16_LE);
    snd_pcm_hw_params_set_rate_near(playback_handle, hw_params, 44100, 0);
    snd_pcm_hw_params_set_channels(playback_handle, hw_params, 2);
    snd_pcm_hw_params_set_periods(playback_handle, hw_params, 2, 0);
    snd_pcm_hw_params_set_period_size(playback_handle, hw_params, BUFSIZE, 0);
    snd_pcm_hw_params(playback_handle, hw_params);
    snd_pcm_sw_params_alloca(&sw_params);
    snd_pcm_sw_params_current(playback_handle, sw_params);
    snd_pcm_sw_params_set_avail_min(playback_handle, sw_params, BUFSIZE);
    snd_pcm_sw_params(playback_handle, sw_params);
    return(playback_handle);
}*/

double envelope(int *note_active, int gate, double *env_level, double t, double attack, double decay, double sustain, double release) {

    if (gate)  {
        if (t > attack + decay) return(*env_level = sustain);
        if (t > attack) return(*env_level = 1.0 - (1.0 - sustain) * (t - attack) / decay);
        return(*env_level = t / attack);
    } else {
        if (t > release) {
            if (note_active) *note_active = 0;
            return(*env_level = 0);
        }
        return(*env_level * (1.0 - t / release));
    }
}

int midi_callback() {

    snd_seq_event_t *ev;
    int l1;
  
    do {
        snd_seq_event_input(seq_handle, &ev);
        switch (ev->type) {
            case SND_SEQ_EVENT_PITCHBEND:
                pitch = (double)ev->data.control.value / 8192.0;
                break;
            case SND_SEQ_EVENT_CONTROLLER:
                if (ev->data.control.param == 1) {
                    modulation = (double)ev->data.control.value / 10.0;
                } 
                break;
            case SND_SEQ_EVENT_NOTEON:
                for (l1 = 0; l1 < POLY; l1++) {
                    if (!note_active[l1]) {
                        note[l1] = ev->data.note.note;
                        velocity[l1] = ev->data.note.velocity / 127.0;
                        env_time[l1] = 0;
                        gate[l1] = 1;
                        note_active[l1] = 1;
                        break;
                    }
                }
                break;        
            case SND_SEQ_EVENT_NOTEOFF:
                for (l1 = 0; l1 < POLY; l1++) {
                    if (gate[l1] && note_active[l1] && (note[l1] == ev->data.note.note)) {
                        env_time[l1] = 0;
                        gate[l1] = 0;
                    }
                }
                break;        
        }
        snd_seq_free_event(ev);
    } while (snd_seq_event_input_pending(seq_handle, 0) > 0);
    return (0);
}

int
sample_rate_cb (jack_nframes_t sr, void * data)
{
  sample_rate = sr;
  
  return 0;
}

int
process (jack_nframes_t nframes, void * data)
{
    
    int l1, l2;
    double dphi, dphi_mod, f1, f2, f3, freq_note, sound;
    double * jack_buffer;
    
    jack_buffer = jack_port_get_buffer (jack_port, nframes);
      
    memset(jack_buffer, 0, nframes * sizeof (jack_default_audio_sample_t));
    for (l2 = 0; l2 < POLY; l2++) {
        if (note_active[l2]) {
            f1 = 8.176 * exp((double)(transpose+note[l2]-2)*log(2.0)/12.0);
            f2 = 8.176 * exp((double)(transpose+note[l2])*log(2.0)/12.0);
            f3 = 8.176 * exp((double)(transpose+note[l2]+2)*log(2.0)/12.0);
            freq_note = (pitch > 0) ? f2 + (f3-f2)*pitch : f2 + (f2-f1)*pitch;
            dphi = M_PI * freq_note / (sample_rate/2);                                    
            dphi_mod = dphi * (double)harmonic / (double)subharmonic;
            for (l1 = 0; l1 < nframes; l1++) {
                phi[l2] += dphi;
                phi_mod[l2] += dphi_mod;
                if (phi[l2] > 2.0 * M_PI) phi[l2] -= 2.0 * M_PI;
                if (phi_mod[l2] > 2.0 * M_PI) phi_mod[l2] -= 2.0 * M_PI;
                sound = envelope(&note_active[l2], gate[l2], &env_level[l2], env_time[l2], attack, decay, sustain, release)
                             * velocity[l2] * sin(phi[l2] +  modulation * sin(phi_mod[l2]));
                env_time[l2] += 1.0 / sample_rate;
/*                buf[2 * l1] += sound;
                buf[2 * l1 + 1] += sound;*/
                jack_buffer[l1] = sound;
            }
        }    
    }
/*    return snd_pcm_writei (playback_handle, buf, nframes);  */
  return 0;
}
      
int
synth_main ()
{
    int seq_nfds, l1;
    struct pollfd *pfds;
    
    pitch = 0;
    for (l1 = 0; l1 < POLY; note_active[l1++] = 0);

    open_jack ();
    seq_handle = open_seq ();

    seq_nfds = snd_seq_poll_descriptors_count(seq_handle, POLLIN);
    pfds = (struct pollfd *)alloca(sizeof(struct pollfd) * seq_nfds);
    snd_seq_poll_descriptors(seq_handle, pfds, seq_nfds, POLLIN);
    
    
    while (1) {
        if (poll (pfds, seq_nfds, 1000) > 0) {
            for (l1 = 0; l1 < seq_nfds; l1++) {
               if (pfds[l1].revents > 0) midi_callback();
            }
        }

#ifdef HAVE_LADCCA        
        if (!cca_server_connected (cca_client))
          break;
#endif
    }

    snd_seq_close (seq_handle);
    close_jack ();
    return (0);
}

